import pickle, os
import logging, logging.handlers
import socketserver, struct

class LogRecordStreamHandler(socketserver.StreamRequestHandler):
    logfile = '/tmp/pyliveupdate.log'
    def handle(self):
        """
        Handle multiple requests - each expected to be a 4-byte length,
        followed by the LogRecord in pickle format. Logs the record
        according to whatever policy is configured locally.
        """
        while True:
            chunk = self.connection.recv(4)
            if len(chunk) < 4:
                break
            slen = struct.unpack('>L', chunk)[0]
            chunk = self.connection.recv(slen)
            while len(chunk) < slen:
                chunk = chunk + self.connection.recv(slen - len(chunk))
            obj = self.unPickle(chunk)
            record = logging.makeLogRecord(obj)
#             print('received', record)
            self.handleLogRecord(record)

    def unPickle(self, data):
        return pickle.loads(data)

    def handleLogRecord(self, record):
        # if a name is specified, we use the named logger rather than the one
        # implied by the record.
        if self.server.logname is not None:
            name = self.server.logname
        else:
            name = record.name
        logger = logging.getLogger(name)
        
        if (logger.handlers == []):
            f_handler = logging.FileHandler(self.logfile)
            f_handler.setLevel(logging.INFO)
#             f_format = logging.Formatter('''%(asctime)s - %(name)s - %(levelname)s\n\
#                 - %(processName)s - %(process)s - %(threadName)s\n\
#                 - %(pathname)s - %(module)s - %(funcName)s - %(lineno)s\n\
#                 - %(message)s\n''')
            f_format =logging.Formatter(
                '%(asctime)s, %(processName)s, %(process)s, %(threadName)s, '+\
                '%(pathname)s, %(lineno)s, %(module)s, %(funcName)s, %(message)s\n')
            f_handler.setFormatter(f_format)
            logger.addHandler(f_handler)
        
        # N.B. EVERY record gets logged. This is because Logger.handle
        # is normally called AFTER logger-level filtering. If you want
        # to do filtering, do it at the client end to save wasting
        # cycles and network bandwidth!
        logger.handle(record)

class LogRecordSocketReceiver(socketserver.ThreadingTCPServer):
    allow_reuse_address = True

    def __init__(self, host, port, handler):
        socketserver.ThreadingTCPServer.__init__(self, (host, port), handler)
        self.abort = 0
        self.timeout = 1
        self.logname = None

    def serve_until_stopped(self):
        import select
        abort = 0
        while not abort:
            rd, wr, ex = select.select([self.socket.fileno()],
                                       [], [], self.timeout)
            if rd:
                self.handle_request()
            abort = self.abort

def start_logger(host='localhost',
                 port=logging.handlers.DEFAULT_TCP_LOGGING_PORT, logfile='/tmp/pyliveupdate.log'):
    
    LogRecordStreamHandler.logfile = logfile
    tcpserver = LogRecordSocketReceiver(host, port, LogRecordStreamHandler)
    print('Start logging server: process {} listens on {}:{} '.format(os.getpid(), host, port))

    tcpserver.serve_until_stopped()



import sys
import socket
import traceback
import threading
from code import InteractiveConsole
from io import StringIO
from pyliveupdate import *

class DistantInteractiveConsole(InteractiveConsole):
    def __init__(self, ipc):
        InteractiveConsole.__init__(self, globals())
        self.ipc = ipc
        self.set_buffer()

    def set_buffer(self):
        self.out_buffer = StringIO()
        sys.stdout = sys.stderr = self.out_buffer

    def unset_buffer(self):
        sys.stdout, sys.stderr = sys.__stdout__, sys.__stderr__
        value = self.out_buffer.getvalue()
        self.out_buffer.close()

        return value

    def raw_input(self, prompt=""):
        output = self.unset_buffer()        
        self.ipc.send('\n'.join((prompt, output)))

        try:
            cmd = self.ipc.recv()
        except Exception as e:
            print("reverse Error: ",e)
        
        self.set_buffer()

        return cmd


class UpdateClient(threading.Thread):
    """
    A reverse Python shell that behaves like Python interactive interpreter.
    """
    def __init__(self, host='localhost', port=9001):
        super(UpdateClient, self).__init__()
        self.daemon = True
        self.host = host
        self.port = port
        self.sock = None
    def run(self):
        try:
            for res in socket.getaddrinfo(self.host, self.port,
                    socket.AF_UNSPEC, socket.SOCK_STREAM):
                af, socktype, proto, canonname, sa = res
                try:
                    self.sock = socket.socket(af, socktype, proto)
                    try:
                        self.sock.connect(sa)
                    except socket.error:
                        self.sock.close()
                        self.sock = None
                        continue
                except socket.error:
                    self.sock = None
                    continue
                break

            if not self.sock:
                raise Exception('pyrasite cannot establish reverse ' +
                        'connection to %s:%d' % (self.host, self.port))
            try:
                print('========Update client started=========')
                ipc = IPC(self.sock)
                DistantInteractiveConsole(ipc).interact()
            except Exception as e:
                print(e)
            finally:
                self.sock.close()
        except SystemExit:
            pass
        except:
            traceback.print_exc(file=sys.__stderr__)
        finally:
            sys.stdout, sys.stderr = sys.__stdout__, sys.__stderr__


from pyliveupdate.controller import UpdateServer, start_logger
import argparse, multiprocessing, os
from pyliveupdatescripts import *

def main():
    parser = argparse.ArgumentParser(description='Update server')
    parser.add_argument('-l', '--logserver', 
                        help='start log server to accept logs from client',
                        action="store_true")
    parser.add_argument('--attachscript', type=str, 
                        help='an update script to attach to the target program')
    args = parser.parse_args()
    
    if args.logserver:
        multiprocessing.Process(target=start_logger).start()
    
    startpayloads = ['exec("from pyliveupdatescripts import *")']
    
    if args.attachscript:
        startpayloads.append(r"exec(r'''{}''')".format(open(args.attachscript).read()))
    controller = UpdateServer()
    controller.start(startpayloads)
    
if __name__ == '__main__':
    main()


from pyliveupdate import *
import time, inspect

def func_begin():
    import time
    global start
    start = time.time()
    #print(time())
    
def func_end():
    from pyliveupdate import client
    import time, inspect
    global start
#     code_ = inspect.currentframe().f_code
#     func_name = '{} {}'.format(code_.co_filename, code_.co_name)
    time_ = (time.time()-start)*1000
#     print('{} execute time: {:.6f} ms'.format(func_name, time_))
    client.local_logger.info('execution time: {:.4f} ms'.format(time_))
    client.remote_logger.info('execution time: {:.4f} ms'.format(time_))

def instru(scope):
    update = Update('instrument', scope, 
                    {'func_begin':func_begin, 'func_end':func_end},
                   None, None, __name__)
    UpdateManager.apply_update(update)
    
def revert(updateid):
    UpdateManager.revert_update(updateid)

def listinstru():
    print(UpdateManager.updates)


import logging, logging.handlers, queue, time, os

class mysocketHandler(logging.handlers.SocketHandler):
    def emit(self, record):
        try:
            s = self.makePickle(record)
            print(s)
            self.send(s)
        except Exception:
            self.handleError(record)
            
def get_remote_logger(host='localhost', port=logging.handlers.DEFAULT_TCP_LOGGING_PORT):
    que = queue.Queue(-1)  # no limit on size
    queue_handler = logging.handlers.QueueHandler(que)

    socketHandler = logging.handlers.SocketHandler(host, port)
#     socketHandler = mysocketHandler(host, port)
    socketHandler.setLevel(logging.INFO)
#     c_format = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
#     socketHandler.setFormatter(c_format)

    listener = logging.handlers.QueueListener(que, socketHandler)
    listener.start()
    
    logger = logging.getLogger(__name__+'.remote_logger')
    logger.setLevel(logging.INFO)
    logger.addHandler(queue_handler)
    return logger

def get_local_logger():
    logger = logging.getLogger(__name__+'.local_logger')
    c_handler = logging.StreamHandler()
    c_handler.setLevel(logging.INFO)
    c_format =logging.Formatter(
#                 '%(asctime)s - %(processName)s - %(process)s - %(threadName)s -'+\
                '%(pathname)s, %(lineno)s, %(module)s.%(funcName)s: %(message)s\n')
    c_handler.setFormatter(c_format)
    logger.setLevel(logging.INFO)
    logger.addHandler(c_handler)
    return logger

remote_logger = get_remote_logger()
local_logger = get_local_logger()


#!/usr/bin/python
# coding=utf8
"""
# Author: lily
# Created Time : 2020-02-05 15:05:54

# File Name: register.py
# Description:

"""
def plu_register(reg_list):
    g_list = reg_list
    for item in g_list:
        if item == '__builtins__':
            continue
        print(item)
        if __builtins__.get(item) == None and item[:2] != '__' and item != 'plu_register':
            print('ADD TO BUILTIN:', item)
            __builtins__[item] = g_list[item]





from pyliveupdate import *
from pyliveupdatescripts import *
import time

class FuncProfiler(UpdateBase):
    start = 0
    
    @staticmethod
    def _func_begin():
        FuncProfiler.start = time.time()
    
    @staticmethod
    def _func_end():
        time_ = (time.time()-FuncProfiler.start)*1000
        local_logger.info('execution time: {:.4f} ms'.format(time_))
        remote_logger.info('execution time: {:.4f} ms'.format(time_))
    
    @staticmethod
    def profile(scope):
        scopes = scope
        if isinstance(scope, str):
            scopes = [scope]
        for scope in scopes:
            update = Update('instrument', scope, 
                            {'func_begin': FuncProfiler._func_begin, 
                             'func_end': FuncProfiler._func_end},
                           None, None, __name__)
            UpdateManager.apply_update(update)

FP = FuncProfiler
FuncProfiler.register_builtin(globals())


import struct

class IPC(object):
    def __init__(self, sock):
        super(IPC, self).__init__()
        self.sock = sock

    def send(self, data):
        """
        Send arbitrary data to the process via self.sock
        """
        header = ''.encode('utf-8')
        data = data.encode('utf-8')
        header = struct.pack('<L', len(data))
        self.sock.sendall(header + data)

    def recv(self):
        """
        Receive a command from a given socket
        """
        header_data = self.recv_bytes(4)
        
        if len(header_data) == 4:
            msg_len = struct.unpack('<L', header_data)[0]
#             print("header_data: ", struct.unpack('<L', header_data))
            data = self.recv_bytes(msg_len).decode('utf-8')
#             print("ipc.recv.data : ", data, "|")
            if len(data) == msg_len:
                return data

    def recv_bytes(self, n):
        """
        Receive n bytes from a socket
        """
        data = ''.encode('utf-8')
        while len(data) < n:
            chunk = self.sock.recv(n - len(data))
            if not chunk:
                break
            data += chunk
        return data

    def __repr__(self):
        return "<%s %s>" % (self.__class__.__name__, self.pid)



from pyliveupdate import IPC
import os, sys, stat, socket, struct, readline
import tempfile, traceback, subprocess, platform
from os.path import dirname, abspath, join


class UpdateServer():
    def __init__(self, host='localhost', port=9001, timeout=5):
        super(UpdateServer, self).__init__()
        self.sock = None
        self.server_sock = None
        self.hostname = host
        self.port = port
        self.timeout = float(timeout)

    def __exit__(self, *args, **kwargs):
        try:
            if self.sock:
                self.sock.close()
            if self.server_sock:
                self.server_sock.close()
        except:
            traceback.print_exc()

    def _listen(self):
        """Listen on a random port"""
        for res in socket.getaddrinfo('localhost', 9001, socket.AF_UNSPEC,
                                      socket.SOCK_STREAM, 0, 0):
            af, socktype, proto, canonname, sa = res
            try:
                self.server_sock = socket.socket(af, socktype, proto)
                try:
                    self.server_sock.bind(sa)
                    self.server_sock.listen(1)
                except socket.error:
                    self.server_sock.close()
                    self.server_sock = None
                    continue
            except socket.error:
                print("error 117")
                self.server_sock = None
                continue
            break

        if not self.server_sock:
            raise Exception('Unable to setup a local server socket')
        else:
            self.hostname, self.port = self.server_sock.getsockname()[0:2]
            print("Start update server starts: process {} listening on {}:{}".format(
                os.getpid(), self.hostname, self.port))

    def _wait(self):
        """Wait for the injected payload to connect back to us"""
        (clientsocket, address) = self.server_sock.accept()
        self.sock = clientsocket
        #print("ipc.wait. self.sock: ", self.sock)
        self.sock.settimeout(self.timeout)
        self.address = address
        
    def start(self):
#         print('Starting update server')
        self._listen()
        self._wait()
        ipc = IPC(self.sock)
        prompt, payload = ipc.recv().split('\n', 1)
        
        ### py3k compat ###
        try:
            input_ = raw_input
        except NameError:
            input_ = input
        
        while True:
            try:
                try:
                    input_line = input_(prompt)
                except Exception as e:
                    print("Error: ", e, prompt)
                    input_line = '101 exit()'
                    ipc.send(input_line)
                    break
                ipc.send(input_line)
                payload = ipc.recv()
                if payload is not None:
                    prompt, payload = payload.split('\n', 1)
                    print(payload)
            except:
                raise Exception('Server error')


import inspect, sys, time
from bytecode import UNSET, Label, Instr, Bytecode, BasicBlock, ControlFlowGraph
import copy, types, os
import pathlib

def get_builtin_list():
    from stdlib_list import stdlib_list
    import sys, os
    import distutils.sysconfig as sysconfig
    
    builtin_list = stdlib_list()+dir(sys.modules['builtins'])
    
    std_lib_path = sysconfig.get_python_lib(standard_lib=True)
    for fname in os.listdir(std_lib_path):
        if fname[-3:] == '.py':
            fname = fname[:-3]
        builtin_list.append(fname)
    for fname in os.listdir(std_lib_path+'/site-packages'):
        if fname[-3:] == '.py':
            fname = fname[:-3]
        builtin_list.append(fname)
    builtin_list += ['pyliveupdate', 'bytecode']
    return list(set(builtin_list))

class TargetFilter(object):
    builtin_list = get_builtin_list()
    
    @classmethod
    def filter_builtin(cls, name, obj):
        '''check if an obj with name is a builtin type
        or and object of a builtin type
        '''
        name = name.split('.')[0]
        if name == '__main__':
            return False
        if name.startswith('_'):
            return True
        elif inspect.isbuiltin(obj):
            return True
        elif inspect.ismodule(obj):
            return (name in cls.builtin_list)
        elif inspect.isclass(obj) or inspect.ismethod(obj) \
            or inspect.isfunction(obj):
            modname = obj.__module__.split('.')[0]
            if modname == '__main__':
                return False
            else:
                return (name in cls.builtin_list)\
                    or (modname in cls.builtin_list)     
        else:
            typename = type(obj).__qualname__.split('.')[0]
            return typename in cls.builtin_list
        
    @classmethod
    def filter_target(cls, name, obj, target):
        '''check if an obj is in an module under the target directory
        '''
        objmod = obj
        if inspect.isclass(obj) or inspect.ismethod(obj) \
            or inspect.isfunction(obj):
            
            if obj.__module__ in sys.modules:
                objmod = sys.modules[obj.__module__]
            
        elif inspect.ismodule(obj):
            objmod = obj
        else:
            objmod = sys.modules[type(obj).__module__]
        
        ### keep __main__ ###
        if objmod.__name__ == '__main__':
            return True
        ### builtin moduel doesn't have __file__###
        elif hasattr(objmod, '__file__') and objmod.__file__\
            and os.path.realpath(objmod.__file__).startswith(target):
            #print(obj, os.path.realpath(objmod.__file__), target)
            return True
        else:
            return False
        

class TargetFinder(object):
    @classmethod
    def _find_target_rec(cls, path, name_candidates, handled_candidates, target_dir):
        if len(path) <= 0 or len(name_candidates) <= 0:
            return []
        result = []
        new_candidates = []
        p = path[0]
        path2 = path[1:]
        
        ### all matched ###
        if p == '*' or p == '**':
            for name, candidate in name_candidates:
                if id(candidate) not in handled_candidates\
                    and TargetFilter.filter_target(name, candidate, target_dir):
        
                    handled_candidates.add(id(candidate))
                    for new_name, new_candidate in inspect.getmembers(candidate):
                        new_candidates.append((new_name, new_candidate))
                        
                    if len(path) == 1:
                        result.append(candidate)
            if p == '**':
                path2 = ['**']
        else:
            for name, candidate in name_candidates:
                if p == name:
                    for new_name, new_candidate in inspect.getmembers(candidate):
                        new_candidates.append((new_name, new_candidate))
                    if len(path) == 1:
                        result.append(candidate)
        result.extend(TargetFinder._find_target_rec(
                      path2, new_candidates, handled_candidates, target_dir))
        return result

    @classmethod
    def find(cls, target_name, bypass_module):
        '''
        find a target object (function, class, etc,) 
        with target_name (such as '__main__.foo') in sys.modules
        '''
        modules = sys.modules
        path = target_name.split('.')
        result = []
        if len(path) <= 0:
            return result
        else:
            ### '**' and '*' match with all targets in current work directory
            if path[0] == '**' or path[0] == '*':
                #target_dir = str(pathlib.Path(__file__).parent.absolute())
                target_dir = os.getcwd()

            ### match with targets under given module path
            elif path[0] in sys.modules and sys.modules[path[0]].__file__:
                #target_dir = os.path.dirname(sys.modules[path[0]].__file__)
                target_dir = str(pathlib.Path(sys.modules[path[0]].__file__))
            else:
                return result
            new_candidates = [(n, m) for n, m in modules.items() if n != bypass_module]
            result.extend(TargetFinder._find_target_rec(path, new_candidates, set(), target_dir))

        return result

class Instrumenter(object):
    @staticmethod
    def _extract_payload(payload_func):
        ### -2 means remove return instr from the payload function
        return Bytecode.from_code(payload_func.__code__)[:-2]
    

    def _set_payload_lineno(payload_bytecode, lineno):
        new_bytecode = []
        for bc in payload_bytecode:
            new_bytecode.append(Instr(bc.name, bc.arg))
            new_bytecode[-1].lineno = lineno
        return new_bytecode

    def _set_lineno_var(payload_bytecode, lineno):
        payload_bytecode[0:0] = [Instr('LOAD_CONST', lineno), Instr('STORE_FAST', '_watch_lineno')]
        payload_bytecode[0].lineno = lineno
        payload_bytecode[1].lineno = lineno

    def _set_watchvar_var(payload_bytecode, var_name, lineno):
        payload_bytecode[0:0] = [Instr('LOAD_FAST', var_name), Instr('STORE_FAST', '_watch_variable')]
        payload_bytecode[0].lineno = lineno
        payload_bytecode[1].lineno = lineno
 
    def _substitute_var(payload_bytecode, var_name, lineno):
        new_bytecode = []
        for bc in payload_bytecode:
            new_bytecode.append(Instr(bc.name, bc.arg))
            new_bytecode[-1].lineno = lineno
        Instrumenter._set_lineno_var(new_bytecode, lineno)
        Instrumenter._set_watchvar_var(new_bytecode, var_name, lineno)
        return new_bytecode

           
    @staticmethod
    def _instrument_func_begin(func_bytecode, payload_bytecode):
        beginpoint = 0
        lineno = func_bytecode[beginpoint].lineno
        payload_bytecode = Instrumenter._set_payload_lineno(payload_bytecode, lineno)
        func_bytecode[beginpoint:beginpoint] = payload_bytecode
        return func_bytecode

  
    @staticmethod
    def find_func_end(func_bytecode):
        endpoints = []
        for i, instr in enumerate(func_bytecode):
            if isinstance(instr, Instr) and instr.name == 'RETURN_VALUE':
                endpoints.append(i)
        return endpoints

    @staticmethod
    def find_watch_pt(func_bytecode, var_name):
        watchpoints = []
        for i, instr in enumerate(func_bytecode):
            if isinstance(instr, Instr) and instr.name == 'STORE_FAST' and instr.arg == var_name:
                watchpoints.append(i + 1)
        return watchpoints

    @staticmethod
    def _instrument_func_end(func_bytecode, payload_bytecode):
        for endpoint in reversed(Instrumenter.find_func_end(func_bytecode)):
            lineno = func_bytecode[endpoint].lineno
            payload_bytecode = Instrumenter._set_payload_lineno(
                payload_bytecode, lineno)
            func_bytecode[endpoint:endpoint] = payload_bytecode
        return func_bytecode

    @staticmethod
    def find_line_begin(func_bytecode):
        endpoints = []
        last_lineno = -1
        for i, instr in enumerate(func_bytecode):
            if instr.lineno != last_lineno:
                last_lineno = instr.lineno
                endpoints.append(i)
        return endpoints

    @staticmethod
    def _instrument_line_begin(func_bytecode, payload_bytecode):
        endpoint_list = list(reversed(Instrumenter.find_line_begin(func_bytecode)))
        for endpoint in endpoint_list:
            lineno = func_bytecode[endpoint].lineno
            new_bytecode = Instrumenter._set_payload_lineno(
                payload_bytecode, lineno)
            Instrumenter._set_lineno_var(new_bytecode, lineno)
            func_bytecode[endpoint:endpoint] = new_bytecode
        return func_bytecode

    @staticmethod
    def find_line_end(func_bytecode):
        endpoints = []
        last_lineno = -1
        rev_func_bytecode = reversed(func_bytecode)
        func_len = len(func_bytecode)
        for i, instr in enumerate(rev_func_bytecode):
            if instr.lineno != last_lineno:
                last_lineno = instr.lineno
                if instr.name != 'RETURN_VALUE':
                    endpoints.append(func_len - i)
                else:
                    endpoints.append(func_len - i - 1)
        return endpoints

    @staticmethod
    def _instrument_line_end(func_bytecode, payload_bytecode):
        endpoint_list = Instrumenter.find_line_end(func_bytecode)
        for endpoint in endpoint_list:
            lineno = func_bytecode[endpoint-1].lineno
            new_bytecode = Instrumenter._set_payload_lineno(
                payload_bytecode, lineno)
            Instrumenter._set_lineno_var(new_bytecode, lineno)
            func_bytecode[endpoint:endpoint] = new_bytecode
        return func_bytecode


    @staticmethod
    def _instrument_variable_watchpt(func_bytecode, payload_bytecode, var_name):
        watchpt_list = reversed(Instrumenter.find_watch_pt(func_bytecode, var_name))
        for watchpt in watchpt_list:
            lineno = func_bytecode[watchpt-1].lineno
            new_bytecode = Instrumenter._substitute_var(payload_bytecode, var_name, lineno)
            func_bytecode[watchpt:watchpt] = new_bytecode
        return func_bytecode

    @staticmethod
    def instrument_function(target, jointpoint_payload, variable = None):
        func_bytecode = Bytecode.from_code(target.__code__)
        beginpoint = 0
        endpoint = -2
        for jointpoint, payload_func in jointpoint_payload.items():
            payload_bytecode = Instrumenter._extract_payload(payload_func)
            if jointpoint == 'func_begin':
                func_bytecode = Instrumenter._instrument_func_begin(func_bytecode, payload_bytecode)
            elif jointpoint == 'func_end':
                func_bytecode = Instrumenter._instrument_func_end(func_bytecode, payload_bytecode)
            elif jointpoint == 'watch_var':
                var_name = variable
                func_bytecode = Instrumenter._instrument_variable_watchpt(func_bytecode, payload_bytecode, var_name)
            elif jointpoint == 'line_begin':
                func_bytecode = Instrumenter._instrument_line_begin(func_bytecode, payload_bytecode)
            elif jointpoint == 'line_end':
                func_bytecode = Instrumenter._instrument_line_end(func_bytecode, payload_bytecode)
            else:
                #ToDo
                pass
        target.__code__ = func_bytecode.to_code()
                
class Update(object):
    '''
    define an update
    '''
    cnt = 0
    def __init__(self, update_type, target_name, jointpoint_payload, 
                 update_trigger = None, revert_trigger=None, bypass_module=None, variable = None):
        '''
        update_type: instrument or replace
        jointpoint_payload: {jointpoint: payload}, such as {'func_begin': func_begin}
        jointpoint: which code point to apply update, like 'function begin', 'function end'
        payload: a function object whose body will be injected to the jointpoint
        update_trigger: when to trigger the update
        revert_trigger: when to revert the update
        '''
        Update.cnt += 1
        self.id = Update.cnt
        self.type = update_type
        self.target_name = target_name
        self.jointpoint_payload = jointpoint_payload
        self.update_trigger = update_trigger  ## ToDo
        self.revert_trigger = revert_trigger  ## ToDo
        
        self.targets = TargetFinder.find(target_name, bypass_module)
        self.old_codes = {}  ### keep records of the old target code
        self.applied = False ### keep track if the update has been applied or not
        self.variable = variable
    
    def apply(self):
        if self.applied:
            return False
        else:
            if self.type == 'instrument':
                updated_targets = []
                for target in self.targets:
                    if inspect.isfunction(target) or inspect.ismethod(target):
                        target_func = target
                        if inspect.ismethod(target):
                            target_func = target.__func__
                            
                        if target_func not in updated_targets \
                               and hasattr(target_func, '__code__'):
                            codeobj = target_func.__code__
                            Instrumenter.instrument_function(target_func, 
                                                             self.jointpoint_payload, self.variable)
                            updated_targets.append(target_func)
                            self.old_codes[target_func] = codeobj
                    else:
                        #ToDo
                        pass 
                else:
                    #ToDo
                    pass
            self.applied = True
            return True
        
    def revert(self):
        for target, codeobj in self.old_codes.items():
            if inspect.isfunction(target) and hasattr(target, '__code__'):
                target.__code__ = codeobj
            elif inspect.ismethod(target) and hasattr(target.__func__, '__code__'):
                target.__func__.__code__ = codeobj
        self.applied = False
    
    def __repr__(self):
        if self.applied:
            status = 'applied'
        else:
            status = 'not applied'
        return '{}, {}, {}, {}'.format(status, self.type, self.target_name, self.targets)

class UpdateManager(object):
    updates = {}  #{update_id: update}
    
    @classmethod
    def add_update(cls, update):
        cls.updates[update.id] = update
        
    @classmethod
    def rm_update(cls, updateid):
        if updateid in cls.updates:
            del cls.updates[updateid]
    
    @classmethod
    def apply_update(cls, update_or_id):
        updateid = update_or_id
        if isinstance(update_or_id, Update):
            cls.add_update(update_or_id)
            updateid = update_or_id.id
            
        if updateid in cls.updates:
            update = cls.updates[updateid]
            update.apply()
            print('Update {} applied'.format(updateid))
        else:
            print('Update {} not find'.format(updateid))
        
    @classmethod
    def revert_update(cls, updateid):
        if updateid in cls.updates:
            cls.updates[updateid].revert()
            print('Update {} reverted'.format(updateid))
        else:
            print('Update {} not found'.format(updateid))
            


from pyliveupdate.server import UpdateServer, start_logger
import argparse, multiprocessing, os

def main():
    parser = argparse.ArgumentParser(description='Update server')
    parser.add_argument('-l', 
                        '--logserver', 
                        help='start log server to accept logs from client',
                        action="store_true")
    args = parser.parse_args()
    if args.logserver:
        multiprocessing.Process(target=start_logger).start()
    UpdateServer().start()
    
if __name__ == '__main__':
    main()


import pickle, os
import logging, logging.handlers
import socketserver, struct

class LogRecordStreamHandler(socketserver.StreamRequestHandler):
    server_logger = logging.getLogger(__name__)
    def handle(self):
        """
        Handle multiple requests - each expected to be a 4-byte length,
        followed by the LogRecord in pickle format. Logs the record
        according to whatever policy is configured locally.
        """
        while True:
            chunk = self.connection.recv(4)
            if len(chunk) < 4:
                break
            slen = struct.unpack('>L', chunk)[0]
            chunk = self.connection.recv(slen)
            while len(chunk) < slen:
                chunk = chunk + self.connection.recv(slen - len(chunk))
            obj = self.unPickle(chunk)
            record = logging.makeLogRecord(obj)
#             print('received', record)
            self.handleLogRecord(record)

    def unPickle(self, data):
        return pickle.loads(data)

    def handleLogRecord(self, record):
        # if a name is specified, we use the named logger rather than the one
        # implied by the record.
#         if self.server.logname is not None:
#             name = self.server.logname
#         else:
#             name = record.name
#         logger = logging.getLogger(name)
        
#         if (logger.handlers == []):
#             # print(logger.name, logger.propagate , logger.handlers )
#             f_handler = logging.FileHandler(self.logfile)
#             f_handler.setLevel(logging.INFO)
# #             f_format = logging.Formatter('''%(asctime)s - %(name)s - %(levelname)s\n\
# #                 - %(processName)s - %(process)s - %(threadName)s\n\
# #                 - %(pathname)s - %(module)s - %(funcName)s - %(lineno)s\n\
# #                 - %(message)s\n''')
#             f_format =logging.Formatter(
#                 '%(asctime)s, %(processName)s, %(process)s, %(threadName)s, '+\
#                 '%(pathname)s, %(lineno)s, %(module)s, %(funcName)s, %(message)s\n')
#             f_handler.setFormatter(f_format)
#             logger.addHandler(f_handler)
        
        # N.B. EVERY record gets logged. This is because Logger.handle
        # is normally called AFTER logger-level filtering. If you want
        # to do filtering, do it at the client end to save wasting
        # cycles and network bandwidth!
        self.server_logger.handle(record)

class LogRecordSocketReceiver(socketserver.ThreadingTCPServer):
    allow_reuse_address = True

    def __init__(self, host, port, handler):
        socketserver.ThreadingTCPServer.__init__(self, (host, port), handler)
        self.abort = 0
        self.timeout = 1
        self.logname = None

    def serve_until_stopped(self):
        import select
        abort = 0
        while not abort:
            rd, wr, ex = select.select([self.socket.fileno()],
                                       [], [], self.timeout)
            if rd:
                self.handle_request()
            abort = self.abort

def start_logger(host='localhost',
                 port=logging.handlers.DEFAULT_TCP_LOGGING_PORT, logfile = '/tmp/instru.log'):
#     logging.basicConfig(
#         format='%(relativeCreated)5d %(name)-15s %(levelname)-8s %(message)s')
    server_logger = logging.getLogger(__name__)
    f_handler = logging.FileHandler(logfile)
    f_handler.setLevel(logging.INFO)
    f_format =logging.Formatter(
        '%(asctime)s, %(processName)s, %(process)s, %(threadName)s, '+\
        '%(pathname)s, %(lineno)s, %(module)s, %(funcName)s, %(message)s\n')
    f_handler.setFormatter(f_format)
    server_logger.addHandler(f_handler)
    LogRecordStreamHandler.server_logger = server_logger
    tcpserver = LogRecordSocketReceiver(host, port, LogRecordStreamHandler)
    print('Start logging server: process {} listens on {}:{} '.format(os.getpid(), host, port))

    tcpserver.serve_until_stopped()



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pyliveupdate", 
    version="0.1.4",
    author="devopspp",
    keywords='runtime instrumentation logging profiling debugging',
    description="A tool to manipulate python code at runtime for logging, profiling, debugging, etc.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/devopspp/pyliveupdate",
    license='GPLv3',
    packages=setuptools.find_packages(exclude=['tests', 'experimental', 'examples']),
    python_requires='>=3.5',
    install_requires=[
        'bytecode',
        'stdlib_list',
      ],
    entry_points="""
          [console_scripts]
          pylu-controller = pyliveupdate.startcontroller:main
      """,
    classifiers=[
          'Development Status :: 3 - Alpha',
          'Environment :: Console',
          'Intended Audience :: Developers',
          'Intended Audience :: System Administrators',
          'License :: OSI Approved :: GNU General Public License (GPL)',
          'Topic :: System :: Logging',
          'Topic :: System :: Monitoring',
          'Topic :: Software Development :: Libraries :: Python Modules',
          'Topic :: Software Development :: Debuggers',
          'Programming Language :: Python :: 3',
      ]
)



from pyliveupdate import *

class UpdateBase():
    @staticmethod
    def apply(updateid):
        UpdateManager.apply_update(updateid)
    
    @staticmethod
    def revert(updateid):
        UpdateManager.revert_update(updateid)
    
    @staticmethod
    def ls():
        for updateid, update in UpdateManager.updates.items():
            print(updateid, ':', update)
    
    @staticmethod
    def register_builtin(vars_):
        '''
        register vars_ to builtin namespace 
        so that the updated target programs can access
        '''
        for item in vars_:
            if __builtins__.get(item) == None and item[:2] != '__':
                # print('ADD TO BUILTIN:', item)
                __builtins__[item] = vars_[item]